# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

Nx = None
dx = None
Lx = None

Ny = None
dy = None
Ly = None

Nz = None
dz = None
Lz = None

NtA = None
dtA = None
TA = None 

material_tiling = None
order_parameter = 'random'
random_seed = None
random_level = 1.0

normal_conductivity = 1.0 
linear_coefficient = 1.0

homogeneous_external_field = 0.0
external_field = 0.0

fixed_vortices = None
fixed_vortices_correction = 'cell centers'
phase_lock_radius = None

order_parameter_Langevin_coefficient = 0.0
vector_potential_Langevin_coefficient = 0.0
